#include "ImpressionistModel.h"
#include "PaintView.h"
#include "AbstractBrush.h"

//==========[ Constant Definitions ]===========================================

#define LEFT_MOUSE_DOWN		1
#define LEFT_MOUSE_DRAG		2
#define LEFT_MOUSE_UP		3
#define RIGHT_MOUSE_DOWN	4
#define RIGHT_MOUSE_DRAG	5
#define RIGHT_MOUSE_UP		6

//==========[ Macro Definitions ]==============================================

#ifndef WIN32
#define min(a, b)	( ( (a)<(b) ) ? (a) : (b) )
#endif

static int eventToDo;
static int isAnEvent=0;
static Point coord;

//==========[ PaintView Methods ]==============================================

//----------[ Constructor ]----------------------------------------------------

PaintView::PaintView( ImpressionistModel* model, int x, int y, int w, int h, const char *l )
			: Fl_Gl_Window(x,y,w,h,l) {
	mModel = model;
}

//----------[ draw ]-----------------------------------------------------------

void PaintView::draw() {

	// should only be executed the first time draw is called
	if( !valid() ) {
		glClearColor( 0.7f, 0.7f, 0.7f, 1.0f );

		// we are using OpenGL as a 2D API, so turn of depth
		glDisable( GL_DEPTH_TEST );

		// set up a simple orthographic view
		ortho();

		// clear the color buffer
		glClear( GL_COLOR_BUFFER_BIT );
	}

	// the pixel storage structure is RGB triples in a packed,
	// row-major array going from left to right and bottom to top.

	int bitmapWidth = mModel->getBitmapWidth();
	int bitmapHeight = mModel->getBitmapHeight();

	int windowWidth = w();
	int windowHeight = h();

	int drawWidth = min( windowWidth, bitmapWidth );
	int drawHeight = min( windowHeight, bitmapHeight );

	int startRow = bitmapHeight - drawHeight;

	mPaintBitstart = mModel->getPaintedBitmap() + 3
					   * ((bitmapWidth * startRow));

	if ( mModel->getPaintedBitmap() && !isAnEvent)
		restoreGLBuffer();

	// handle any pending mouse events
	if ( mModel->getPaintedBitmap() && isAnEvent) {
		isAnEvent=0;

		// calculate the source color and brush target coord
		Point source( coord.x, bitmapHeight - coord.y );
		Point target( coord.x, windowHeight - coord.y );
		unsigned char* pixel = mModel->getOriginalPixel( source );
		Color sourceColor( *pixel, *(pixel+1), *(pixel+2) );

		// handle the event
		switch( eventToDo ) {
			case LEFT_MOUSE_DOWN:
				mModel->getCurrentBrush()->brushBegin( mModel->getBrushSize(), sourceColor, target );
				break;
			case LEFT_MOUSE_DRAG:
				mModel->getCurrentBrush()->brushMove( mModel->getBrushSize(), sourceColor, target );
				break;
			case LEFT_MOUSE_UP:
				mModel->getCurrentBrush()->brushEnd( mModel->getBrushSize(), sourceColor, target );

				// on mouse up, save the contents of the front buffer
				saveGLBuffer();
				restoreGLBuffer();
				break;
		}
	}

	glFlush();
}

//----------[ handle ]---------------------------------------------------------

int PaintView::handle( int event ) {
	switch( event ) {
	case FL_PUSH:
		coord.x = Fl::event_x();
		coord.y = Fl::event_y();
		if (Fl::event_button()>1)
			eventToDo=RIGHT_MOUSE_DOWN;
		else
			eventToDo=LEFT_MOUSE_DOWN;
		isAnEvent=1;
		redraw();
		break;
	case FL_DRAG:
		coord.x = Fl::event_x();
		coord.y = Fl::event_y();
		if (Fl::event_button()>1)
			eventToDo=RIGHT_MOUSE_DRAG;
		else
			eventToDo=LEFT_MOUSE_DRAG;
		isAnEvent=1;
		redraw();
		break;
	case FL_RELEASE:
		coord.x = Fl::event_x();
		coord.y = Fl::event_y();
		if (Fl::event_button()>1)
			eventToDo=RIGHT_MOUSE_UP;
		else
			eventToDo=LEFT_MOUSE_UP;
		isAnEvent=1;
		redraw();
	default:
		break;
	}
	return 1;
}

//----------[ refresh ]--------------------------------------------------------

void PaintView::refresh() {
	redraw();
}

//----------[ resizeWindow ]---------------------------------------------------

void PaintView::resizeWindow( int width, int height ) {
	resize( x(), y(), width, height );
}

//----------[ saveGLBuffer ]---------------------------------------------------

void PaintView::saveGLBuffer() {
	glReadBuffer( GL_BACK );
	glPixelStorei( GL_PACK_ALIGNMENT, 1 );
	glPixelStorei( GL_PACK_ROW_LENGTH, mModel->getBitmapWidth() );
	
	int windowWidth = w();
	int windowHeight = h();

	int drawWidth = min( windowWidth, mModel->getBitmapWidth() );
	int drawHeight = min( windowHeight, mModel->getBitmapHeight() );

	glReadPixels( 0, windowHeight - drawHeight, drawWidth, drawHeight, GL_RGB,
				  GL_UNSIGNED_BYTE, mPaintBitstart );
}

//----------[ restoreGLBuffer ]------------------------------------------------

void PaintView::restoreGLBuffer() {
	glDrawBuffer( GL_BACK );

	glClear( GL_COLOR_BUFFER_BIT );

	int windowWidth = w();
	int windowHeight = h();

	int drawWidth = min( windowWidth, mModel->getBitmapWidth() );
	int drawHeight = min( windowHeight, mModel->getBitmapHeight() );

	glRasterPos2i( 0, windowHeight - drawHeight );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	glPixelStorei( GL_UNPACK_ROW_LENGTH, mModel->getBitmapWidth() );
	glDrawPixels( drawWidth, drawHeight, GL_RGB, GL_UNSIGNED_BYTE, mPaintBitstart);
}